/**
   This is the "rendering pipeline".

   This program takes as its input a Scene data structure
   and it produces, as its output, a filled in FrameBuffer
   data structure. The final FrameBuffer holds the rendered
   image of the scene represented by the Scene object.
*/

public class SceneRender
{
   public static void renderScene(Scene scene, FrameBuffer fb)
   {
      Scene scene2 = new Scene(scene);  // Duplicate the Scene object,
                                        // so that we can modify the copy.
   /* NOTE: We could make it the client's responsibility
            to copy the Scene. In the case of OpenGL,
            the client "duplicates" the scene when it
            passes all of the scene information into the
            pipeline, and so all of the information gets
            copied into the graphics card's memory.
   */
      printMatrices(scene2);
      printVertices(scene2, 1, 0);

      P1_Model2World.P1_Model2World(scene2); // apply the model transformation

      printVertices(scene2, 2, 1);

      P2_World2View.P2_World2View(scene2);  // apply the view transformation

      printVertices(scene2, 4, 2);

      P3_BackFaceCulling.P3_BackFaceCulling(scene2); // delete back facing triangles

      printVertices(scene2, 8, 3);

      P4_Lighting.P4_Lighting(scene2);    // compute vertex shading

      printVertices(scene2, 16, 4);

      P5_View2Clip.P5_View2Clip(scene2);  // apply the projection transformation

      printVertices(scene2, 32, 5);

      P6_ClippingSimple.P6_Clipping(scene2); // clip triangles to the standard volume
//    P6_Clipping.P6_Clipping(scene2);       // clip triangles to the standard volume

      printVertices(scene2, 64, 6);

      P7_Clip2NDC.P7_Clip2NDC(scene2);   // transform to normalized device coordinates

      printVertices(scene2, 128, 7);

      P8_NDC2VP.P8_NDC2VP(scene2); // transform normalized device coordinates to viewport coordinates

      printVertices(scene2, 256, 8);

      P9_Rasterize.P9_Rasterize(scene2);  // rasterize each triangle into fragments

      printVertices(scene2, 512, 9);

      P10_FragmentShaderTexture.P10_FragmentShaderTexture(scene2); // shade each fragment of each triangle

      printVertices(scene2, 1024, 10);

      P11_Fragments2FB.P11_Fragments2FB(scene2, fb);  // use the fragments to fill in the viewport
   }//renderScene()


   /**
      Used for debugging.
   */
   public static void printMatrices(Scene scene)
   {
      // Display the matrices stored in the Scene object.
      int printMatrices = 0;
      String matrixString = System.getenv("PrintMatrices");
      if (matrixString != null)
      {
         printMatrices = Integer.parseInt(matrixString); // 1 => print model matrix
      }                                                  // 2 => print view matrix
                                                         // 4 => print projection matrix
      if ((printMatrices & 1) == 1)
      {
         System.err.println( scene.model.toString() );
      }
      if ((printMatrices & 2) == 2)
      {
         System.err.println( scene.view.toString() );
      }
      if ((printMatrices & 4) == 4)
      {
         System.err.println( scene.projection.toString() );
      }
   }//printMatrices()


   /**
      Used for debugging.
   */
   public static void printVertices(Scene scene, int stageFlag, int stageNo)
   {
      int debugStages = 0;
      String debugString = null;
      if ( (debugString = System.getenv("PipeLine0")) != null ) debugStages += 1;
      if ( (debugString = System.getenv("PipeLine1")) != null ) debugStages += 2;
      if ( (debugString = System.getenv("PipeLine2")) != null ) debugStages += 4;
      if ( (debugString = System.getenv("PipeLine3")) != null ) debugStages += 8;
      if ( (debugString = System.getenv("PipeLine4")) != null ) debugStages += 16;
      if ( (debugString = System.getenv("PipeLine5")) != null ) debugStages += 32;
      if ( (debugString = System.getenv("PipeLine6")) != null ) debugStages += 64;
      if ( (debugString = System.getenv("PipeLine7")) != null ) debugStages += 128;
      if ( (debugString = System.getenv("PipeLine8")) != null ) debugStages += 256;
      if ( (debugString = System.getenv("PipeLine9")) != null ) debugStages += 512;
      if ( (debugString = System.getenv("PipeLine10")) != null ) debugStages += 1024;
      if ( (debugString = System.getenv("PipeLine11")) != null ) debugStages += 2048;
      //if (debugStages == 0) debugStages = 128; // default to NDC

      int printFields = 1;  // default to printing out vertex data
      String printString = System.getenv("PrintLevel");
      if (printString != null)
      {
         printFields = Integer.parseInt(printString); // 1 => print vertex data
      }                                               // 2 => print color data
                                                      // 4 => print texture coordinates
                                                      // 8 => print normal vector
      if ((stageFlag & debugStages) != 0)
      {
         System.err.printf("================= Stage P%d begin ===========\n", stageNo);
         System.err.print( scene.toString(printFields) );
         System.err.printf("================= Stage P%d end =============\n", stageNo);
      }
   }//printVertices()
}