/**
   k number of triangles in the triangle fan at the base of the cone
   n number of triangle strips between the top and bottom
   r radius of the base of the cone
   h height of the cone
*/
import java.awt.Color;

public class Cone3D
{
   /**
      A cone with default color (gray) and default material (chrome).
   */
   public static void cone(int k, int n, double radius, double h,
                           Scene scene)
   {
      cone(k, n, radius, h, null, null, scene);
   }


   /**
      A cone with given color and default material (chrome).
   */
   public static void cone(int k, int n, double radius, double h,
                           Color color, Scene scene)
   {
      cone(k, n, radius, h, null, color, scene);
   }


   /**
      A cone with default color (gray) and given material.
   */
   public static void cone(int k, int n, double radius, double h,
                           Material mtrl, Scene scene)
   {
      cone(k, n, radius, h, mtrl, null, scene);
   }


   /**
      A cylinder with given color and given material
      (unless either color or mtrl is null, then use
      either the default color or material).
   */
   public static void cone(int k, int n, double radius, double h,
                           Material mtrl, Color color, Scene scene)
   {
       // Create the cone's geometry.

      // An array of vertices to be used to create triangles.
      Vertex[][] v = new Vertex[n+1][k];

      // Create all the vertices,
      // along with their normal vectors.
      for (int j = 0; j < k; j++)
      {
         double c = Math.cos(j*(2.0*Math.PI)/k);
         double s = Math.sin(j*(2.0*Math.PI)/k);
         for (int i = 0; i <= n; i++)
         {
            double slantRadius = ((double)i/(double)n) * radius;
            v[i][j] = new Vertex( slantRadius*c, h-i*((2.0*h)/n), slantRadius*s );
            // give the vertex a normal vector
            double l = Math.sqrt( radius*radius + h*h );
            (v[i][j]).n = new Vector(c*h/l, radius/l, s*h/l);
         }
      }
      // Vertex at the center of the bottom.
      Vertex bottomCenter = new Vertex(0.0, 0.0, 0.0);
      bottomCenter.n = new Vector(0.0, -1.0, 0.0);


      // Give each vertex the appropriate color.
      int r = 127;  // the default color (gray)
      int g = 127;
      int b = 127;
      int a = 255;
      if ( null != color )  // use the given color
      {
         r = color.getRed();
         g = color.getGreen();
         b = color.getBlue();
         a = color.getAlpha();
      }
      bottomCenter.setColor(r, g, b, a);
      for (int j = 0; j < k; j++)
         for (int i = 0; i <= n; i++)
            v[i][j].setColor(r, g, b, a);


      if ( null == mtrl )
      {
         // Create a default Material object (chrome).
         mtrl = new Material();
         mtrl.ambient[0] = 0.25f;
         mtrl.ambient[1] = 0.25f;
         mtrl.ambient[2] = 0.25f;
         mtrl.ambient[3] = 1.00f;
         mtrl.diffuse[0] = 0.4f;
         mtrl.diffuse[1] = 0.4f;
         mtrl.diffuse[2] = 0.4f;
         mtrl.diffuse[3] = 1.00f;
         mtrl.specular[0] = 0.775f;
         mtrl.specular[1] = 0.775f;
         mtrl.specular[2] = 0.775f;
         mtrl.specular[3] = 1.00f;
         mtrl.s = 76.8f;
      }


      // Create all the triangle strips along the cone wall.
      Triangle t;
      for (int i = 0; i < n; i++)
      {
         for (int j = 0; j < k-1; j++)
         {
            t = new Triangle( v[i][j], v[i][j+1], v[i+1][j+1] );
            t.mtrl = mtrl;
            scene.triangleList.add( t );
            t = new Triangle( v[i+1][j+1], v[i+1][j], v[i][j] );
            t.mtrl = mtrl;
            scene.triangleList.add( t );
         }
         t = new Triangle( v[i][k-1], v[i][0], v[i+1][0] );
         t.mtrl = mtrl;
         scene.triangleList.add( t );
         t = new Triangle( v[i+1][0], v[i+1][k-1], v[i][k-1] );
         t.mtrl = mtrl;
         scene.triangleList.add( t );
      }

      // Create the triangle fan at the bottom.
      // (And we need to change the normal vectors.)
      for (int j = 0; j < k-1; j++)
      {
         v[n][j].n =   new Vector(0.0, -1.0, 0.0);
         v[n][j+1].n = new Vector(0.0, -1.0, 0.0);
         t = new Triangle( bottomCenter, v[n][j], v[n][j+1] );
         t.mtrl = mtrl;
         scene.triangleList.add( t );
      }
      v[n][k-1].n = new Vector(0.0, -1.0, 0.0);
      v[n][0].n   = new Vector(0.0, -1.0, 0.0);
      t = new Triangle( bottomCenter, v[n][k-1], v[n][0] );
      t.mtrl = mtrl;
      scene.triangleList.add( t );
   }
}