/**
                  v[4]
                   +-----------------+ v[5]
                  /|                /|
                /  |              /  |
              /    |            /    |
            /      |          /      |
      v[7] +-----------------+ v[6]  |
           |       |         |       |               y
           |       |         |       |               |
           |       |         |       |               |
           |  v[0] +---------|-------+ v[1]          |
           |      /          |      /                |
           |    /            |    /                  +----. x
           |  /              |  /                   /
           |/                |/                    /
           +-----------------+                    /
          v[3]              v[2]                 z

   x world x-coordinate of v[0]
   y world y-coordinate of v[0]
   z world z-coordinate of v[0]
   xs the size of the cube along the x-axis
   ys the size of the cube along the y-axis
   zs the size of the cube along the z-axis
*/
import java.awt.Color;

public class Cube3D
{
   /**
      A cube with default color (gray) and default material (chrome).
   */
   public static void cube(double x,  double y,  double z,
                           double xs, double ys, double zs,
                           Scene scene)
   {
      cube(x, y, z, xs, ys, zs, null, null, scene);
   }


   /**
      A cube with given color and default material (chrome).
   */
   public static void cube(double x,  double y,  double z,
                           double xs, double ys, double zs,
                           Color color, Scene scene)
   {
      cube(x, y, z, xs, ys, zs, null, color, scene);
   }


   /**
      A cube with default color (gray) and given material.
   */
   public static void cube(double x,  double y,  double z,
                           double xs, double ys, double zs,
                           Material mtrl, Scene scene)
   {
      cube(x, y, z, xs, ys, zs, mtrl, null, scene);
   }


   /**
      A cube with given color and given material
      (unless either color or mtrl is null, then use
      either the default color or material).
   */
   public static void cube(double x,  double y,  double z,
                           double xs, double ys, double zs,
                           Material mtrl, Color color, Scene scene)
   {
      // Create the cube's geometry.
      Vertex[] v = new Vertex[8];
      v[0] = new Vertex(x,    y,    z);      // four vertices around the bottom face
      v[1] = new Vertex(x+xs, y,    z);
      v[2] = new Vertex(x+xs, y,    z+zs);
      v[3] = new Vertex(x,    y,    z+zs);
      v[4] = new Vertex(x,    y+ys, z);      // four vertices around the top face
      v[5] = new Vertex(x+xs, y+ys, z);
      v[6] = new Vertex(x+xs, y+ys, z+zs);
      v[7] = new Vertex(x,    y+ys, z+zs);


      // Give each vertex the appropriate color.
      int r = 127;  // the default color (gray)
      int g = 127;
      int b = 127;
      int a = 255;
      if ( null != color )  // use the given color
      {
         r = color.getRed();
         g = color.getGreen();
         b = color.getBlue();
         a = color.getAlpha();
      }
      for (int i = 0; i < v.length; i++)
         v[i].setColor(r, g, b, a);


      // Create 12 triangles, 2 for each face of the cube.
      Triangle[] t = new Triangle[12];
      Vector n; // used for the normal vectors
      int i = 0;
      // bottom face
      n = new Vector(0.0, -1.0, 0.0);
      v[0].n = v[1].n = v[2].n = v[3].n = n;
      t[i++] = new Triangle(v[0], v[1], v[2]);
      t[i++] = new Triangle(v[2], v[3], v[0]);
      // front face
      n = new Vector(0.0, 0.0, 1.0);
      v[2].n = v[3].n = v[6].n = v[7].n = n;
      t[i++] = new Triangle(v[3], v[2], v[6]);
      t[i++] = new Triangle(v[6], v[7], v[3]);
      // right face
      n = new Vector(1.0, 0.0, 0.0);
      v[1].n = v[2].n = v[5].n = v[6].n = n;
      t[i++] = new Triangle(v[2], v[1], v[5]);
      t[i++] = new Triangle(v[5], v[6], v[2]);
      // back face
      n = new Vector(0.0, 0.0, -1.0);
      v[0].n = v[1].n = v[4].n = v[5].n = n;
      t[i++] = new Triangle(v[1], v[0], v[4]);
      t[i++] = new Triangle(v[4], v[5], v[1]);
      // left face
      n = new Vector(-1.0, 0.0, 0.0);
      v[0].n = v[3].n = v[4].n = v[7].n = n;
      t[i++] = new Triangle(v[0], v[3], v[7]);
      t[i++] = new Triangle(v[7], v[4], v[0]);
      // top face
      n = new Vector(0.0, 1.0, 0.0);
      v[4].n = v[5].n = v[6].n = v[7].n = n;
      t[i++] = new Triangle(v[7], v[6], v[5]);
      t[i++] = new Triangle(v[5], v[4], v[7]);

      if ( null == mtrl )
      {
         // Create a default Material object (chrome).
         mtrl = new Material();
         mtrl.ambient[0] = 0.25f;
         mtrl.ambient[1] = 0.25f;
         mtrl.ambient[2] = 0.25f;
         mtrl.ambient[3] = 1.00f;
         mtrl.diffuse[0] = 0.4f;
         mtrl.diffuse[1] = 0.4f;
         mtrl.diffuse[2] = 0.4f;
         mtrl.diffuse[3] = 1.00f;
         mtrl.specular[0] = 0.775f;
         mtrl.specular[1] = 0.775f;
         mtrl.specular[2] = 0.775f;
         mtrl.specular[3] = 1.00f;
         mtrl.s = 76.8f;
      }

      // Add the new triangles to the scene.
      for (i = 0; i < t.length; i++)
      {
         t[i].mtrl = mtrl;
         scene.triangleList.add( t[i] );
      }
   }
}