/**
   k number of triangles in the triangle fan at the top and bottom
   n number of triangle strips between the top and bottom
   x world x-coordinate of the center of the cylinder
   y world y-coordinate of the center of the cylinder
   z world z-coordinate of the center of the cylinder
   r radius of the cylinder
   h height of the cylinder
*/
import java.awt.Color;

public class Cylinder3D
{
   /**
      A cylinder with default color (gray) and default material (chrome).
   */
   public static void cylinder(int k, int n, double x, double y, double z, double radius, double h,
                               Scene scene)
   {
      cylinder(k, n, x, y, z, radius, h, null, null, scene);
   }


   /**
      A cylinder with given color and default material (chrome).
   */
   public static void cylinder(int k, int n, double x, double y, double z, double radius, double h,
                               Color color, Scene scene)
   {
      cylinder(k, n, x, y, z, radius, h, null, color, scene);
   }


   /**
      A cylinder with default color (gray) and given material.
   */
   public static void cylinder(int k, int n, double x, double y, double z, double radius, double h,
                               Material mtrl, Scene scene)
   {
      cylinder(k, n, x, y, z, radius, h, mtrl, null, scene);
   }


   /**
      A cylinder with given color and given material
      (unless either color or mtrl is null, then use
      either the default color or material).
   */
   public static void cylinder(int k, int n, double x, double y, double z, double radius, double h,
                               Material mtrl, Color color, Scene scene)
   {
       // Create the cylinder's geometry.

      // An array of vertices to be used to create triangles.
      Vertex[][] v = new Vertex[n+1][k];

      // Create all the vertices along the cylinder wall,
      // along with their normal vectors.
      for (int j = 0; j < k; j++)
      {
         double c = Math.cos(j*(2.0*Math.PI)/k);
         double s = Math.sin(j*(2.0*Math.PI)/k);
         for (int i = 0; i <= n; i++)
         {
            v[i][j] = new Vertex( x + radius*c, y + h-i*((2.0*h)/n), z + radius*s );
            // give the vertex a normal vector
            (v[i][j]).n = new Vector(c, 0.0, s);
         }
      }
      // Vertex at the center of the top.
      Vertex topCenter = new Vertex(0.0+x, h+y, 0.0+z);
      topCenter.n = new Vector(0.0, 1.0, 0.0);
      // Vertex at the center of the bottom.
      Vertex bottomCenter = new Vertex(0.0+x, -h+y, 0.0+z);
      bottomCenter.n = new Vector(0.0, -1.0, 0.0);


      // Give each vertex the appropriate color.
      int r = 127;  // the default color (gray)
      int g = 127;
      int b = 127;
      int a = 255;
      if ( null != color )  // use the given color
      {
         r = color.getRed();
         g = color.getGreen();
         b = color.getBlue();
         a = color.getAlpha();
      }
      topCenter.setColor(r, g, b, a);
      bottomCenter.setColor(r, g, b, a);
      for (int j = 0; j < k; j++)
         for (int i = 0; i <= n; i++)
            v[i][j].setColor(r, g, b, a);


      if ( null == mtrl )
      {
         // Create a default Material object (chrome).
         mtrl = new Material();
         mtrl.ambient[0] = 0.25f;
         mtrl.ambient[1] = 0.25f;
         mtrl.ambient[2] = 0.25f;
         mtrl.ambient[3] = 1.00f;
         mtrl.diffuse[0] = 0.4f;
         mtrl.diffuse[1] = 0.4f;
         mtrl.diffuse[2] = 0.4f;
         mtrl.diffuse[3] = 1.00f;
         mtrl.specular[0] = 0.775f;
         mtrl.specular[1] = 0.775f;
         mtrl.specular[2] = 0.775f;
         mtrl.specular[3] = 1.00f;
         mtrl.s = 76.8f;
      }


      // Create all the triangle strips around the cylinder wall.
      Triangle t;
      for (int i = 0; i < n; i++)
      {
         for (int j = 0; j < k-1; j++)
         {
            t = new Triangle( v[i][j], v[i][j+1], v[i+1][j+1] );
            t.mtrl = mtrl;
            scene.triangleList.add( t );
            t = new Triangle( v[i+1][j+1], v[i+1][j], v[i][j] );
            t.mtrl = mtrl;
            scene.triangleList.add( t );
         }
         t = new Triangle( v[i][k-1], v[i][0], v[i+1][0] );
         t.mtrl = mtrl;
         scene.triangleList.add( t );
         t = new Triangle( v[i+1][0], v[i+1][k-1], v[i][k-1] );
         t.mtrl = mtrl;
         scene.triangleList.add( t );
      }


      // Create the triangle fan at the top.
      // (And we need to change the normal vectors.)
      for (int j = 0; j < k-1; j++)
      {
         v[0][j+1].n = new Vector(0.0, 1.0, 0.0);
         v[0][j].n   = new Vector(0.0, 1.0, 0.0);
         t = new Triangle( topCenter, v[0][j+1], v[0][j] );
         t.mtrl = mtrl;
         scene.triangleList.add( t );
      }
      v[0][0].n   = new Vector(0.0, 1.0, 0.0);
      v[0][k-1].n = new Vector(0.0, 1.0, 0.0);
      t = new Triangle( topCenter, v[0][0], v[0][k-1] );
      t.mtrl = mtrl;
      scene.triangleList.add( t );


      // Create the triangle fan at the bottom.
      // (And we need to change the normal vectors.)
      for (int j = 0; j < k-1; j++)
      {
         v[n][j].n =   new Vector(0.0, -1.0, 0.0);
         v[n][j+1].n = new Vector(0.0, -1.0, 0.0);
         t = new Triangle( bottomCenter, v[n][j], v[n][j+1] );
         t.mtrl = mtrl;
         scene.triangleList.add( t );
      }
      v[n][k-1].n = new Vector(0.0, -1.0, 0.0);
      v[n][0].n   = new Vector(0.0, -1.0, 0.0);
      t = new Triangle( bottomCenter, v[n][k-1], v[n][0] );
      t.mtrl = mtrl;
      scene.triangleList.add( t );
   }
}