/**
   k number of triangles in the triangle fan at the top and the bottom
   n number of triangle strips between the top and bottom triangle fans

   x world x-coordinate of the center of the sphere
   y world y-coordinate of the center of the sphere
   z world z-coordinate of the center of the sphere
   r radius of the sphere

   theta1 beginning longitude angle of the section
   theta2 ending longitude angle of the section
   phi1   beginning latitude angle of the section
   phi2   ending latitude angle of the section
*/
import java.awt.Color;

public class SphereSection3D
{

   public static void sphere(int k, int n,
                             double x, double y, double z, double radius,
                             Scene scene)
   {
      sphere(k, n, x, y, z, radius, 0, 360, 0, 180, null, null, scene);
   }


   /**
      A sphere section with default color (gray) and default material (chrome).
   */
   public static void sphere(int k, int n,
                             double x, double y, double z, double radius,
                             double theta1, double theta2, double phi1, double phi2,
                             Scene scene)
   {
      sphere(k, n, x, y, z, radius, theta1, theta2, phi1, phi2, null, null, scene);
   }


   /**
      A sphere section with given color and default material (chrome).
   */
   public static void sphere(int k, int n,
                             double x, double y, double z, double radius,
                             double theta1, double theta2, double phi1, double phi2,
                             Color color, Scene scene)
   {
      sphere(k, n, x, y, z, radius, theta1, theta2, phi1, phi2, null, color, scene);
   }


   /**
      A sphere section with default color (gray) and given material.
   */
   public static void sphere(int k, int n,
                             double x, double y, double z, double radius,
                             double theta1, double theta2, double phi1, double phi2,
                             Material mtrl, Scene scene)
   {
      sphere(k, n, x, y, z, radius, theta1, theta2, phi1, phi2, mtrl, null, scene);
   }


   /**
      A sphere section with given color and given material
      (unless either color or mtrl is null, then use
      either the default color or material).
   */
   public static void sphere(int k, int n,
                             double x, double y, double z, double radius,
                             double theta1, double theta2, double phi1, double phi2,
                             Material mtrl, Color color, Scene scene)
   {
       // Create sphere section's geometry.

      // An array of vertices to be used to create triangles.
      Vertex[][] v = new Vertex[n+1][k];

      // Create all the vertices,
      // along with their normal vectors.
      for (int j = 0; j < k; j++)
      {
         double c1 = Math.cos(theta1 + j*(theta2 - theta1)/k);
         double s1 = Math.sin(theta1 + j*(theta2 - theta1)/k);
         for (int i = 0; i <= n; i++)
         {
            double c2 = Math.cos(phi1 + (i+1)*(phi2 - phi1)/(n+2));
            double s2 = Math.sin(phi1 + (i+1)*(phi2 - phi1)/(n+2));
            v[i][j] = new Vertex(x + radius*s2*c1,
                                 y + radius*c2,
                                 z + radius*s2*s1);
            // give the vertex a normal vector
            (v[i][j]).n = new Vector(s2*c1,
                                     c2,
                                     s2*s1);
         }
      }
      Vertex topPole = new Vertex(0.0+x, radius+y, 0.0+z);
      topPole.n = new Vector(0.0, 1.0, 0.0);
      Vertex bottomPole = new Vertex(0.0+x, -radius+y, 0.0+z);
      bottomPole.n = new Vector(0.0, -1.0, 0.0);


      // Give each vertex the appropriate color.
      int r = 127;  // the default color (gray)
      int g = 127;
      int b = 127;
      int a = 255;
      if ( null != color )  // use the given color
      {
         r = color.getRed();
         g = color.getGreen();
         b = color.getBlue();
         a = color.getAlpha();
      }
      topPole.setColor(r, g, b, a);
      bottomPole.setColor(r, g, b, a);
      for (int j = 0; j < k; j++)
         for (int i = 0; i <= n; i++)
            v[i][j].setColor(r, g, b, a);


      if ( null == mtrl )
      {
         // Create a default Material object (chrome).
         mtrl = new Material();
         mtrl.ambient[0] = 0.25f;
         mtrl.ambient[1] = 0.25f;
         mtrl.ambient[2] = 0.25f;
         mtrl.ambient[3] = 1.00f;
         mtrl.diffuse[0] = 0.4f;
         mtrl.diffuse[1] = 0.4f;
         mtrl.diffuse[2] = 0.4f;
         mtrl.diffuse[3] = 1.00f;
         mtrl.specular[0] = 0.775f;
         mtrl.specular[1] = 0.775f;
         mtrl.specular[2] = 0.775f;
         mtrl.specular[3] = 1.00f;
         mtrl.s = 76.8f;
      }


      // Create the triangle fan at the top.
      Triangle t;
      if (phi1 > 0)
      {
         for (int j = 0; j < k-1; j++)
         {
            t = new Triangle( topPole, v[0][j+1], v[0][j] );
            t.mtrl = mtrl;
            scene.triangleList.add( t );
         }
         t = new Triangle( topPole, v[0][0], v[0][k-1] );
         t.mtrl = mtrl;
         scene.triangleList.add( t );
      }

      // Create all the triangle strips between the top and bottom triangle fans.
      for (int i = 0; i < n; i++)
      {
         for (int j = 0; j < k-1; j++)
         {
            t = new Triangle( v[i][j], v[i][j+1], v[i+1][j+1] );
            t.mtrl = mtrl;
            scene.triangleList.add( t );
            t = new Triangle( v[i+1][j+1], v[i+1][j], v[i][j] );
            t.mtrl = mtrl;
            scene.triangleList.add( t );
         }
         if ( (theta1 == 0) && (theta2 == 360) )
         {
            t = new Triangle( v[i][k-1], v[i][0], v[i+1][0] );
            t.mtrl = mtrl;
            scene.triangleList.add( t );
            t = new Triangle( v[i+1][0], v[i+1][k-1], v[i][k-1] );
            t.mtrl = mtrl;
            scene.triangleList.add( t );
         }
      }


      // Create the triangle fan at the bottom.
      if (phi2 < 180.0)
      {
         for (int j = 0; j < k-1; j++)
         {
            t = new Triangle( bottomPole, v[n][j], v[n][j+1] );
            t.mtrl = mtrl;
            scene.triangleList.add( t );
         }
         t = new Triangle( bottomPole, v[n][k-1], v[n][0] );
         t.mtrl = mtrl;
         scene.triangleList.add( t );
      }
   }
}